﻿// <summary>ソースコード：モデルクラス</summary>
// <author>CommonMP</author>

using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.Diagnostics;

using CommonMP.HYSSOP.CoreImpl.HSData;
using CommonMP.HYSSOP.CoreImpl.HSTools;

using CommonMP.HYMCO.Interface.Model;
using CommonMP.HYMCO.Interface.Data;
using CommonMP.HYMCO.CoreImpl.Data;
using CommonMP.HYMCO.CoreImpl.Tool;



namespace jp.or.jccajsce.UnitGraph
{

    /// <summary><para>class outline:</para>
    /// <para>モデル</para>
    /// </summary>
    /// <remarks><para>history:</para>
    /// <para>[CommonMP][ver 1.0.0][2010/01][新規作成]</para>
    /// </remarks>
    public class UnitGraph : McStateCalModelBase
    {
        //=========================
        // 演算実行前処理関連メソッド
        //=========================

        ///// <summary><para>method outline:</para>
        ///// <para>計算開始時に動作する処理</para>
        ///// </summary>
        ///// <example><para>usage:</para>
        ///// <para>bool bRtn = ReadyCalculation( )</para>
        ///// </example>
        ///// <param name="">無し</param>
        ///// <returns>true :正常 , false:異常</returns>
        ///// <exception cref="">無し</exception>
        ///// <remarks><para>remarks:</para>
        ///// <para>計算開始の　最初に１回だけコールされるメソッドです。不要ならば、本メソッドをオーバーライドする必要はありません。</para>
        ///// </remarks>
        //public override bool ReadyCalculation()

        /// <summary><para>method outline:</para>
        /// <para>入力側の接続情報チェック</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>bool bRtn = ReceiveConnectionCheck(ref csErrorInf)</para>
        /// </example>
        /// <param name="csErrorInf">エラー出力</param>
        /// <param name="lInputDataNum">入力情報数</param>
        /// <param name="csInputCellData">演算に必要な入力情報配列</param>
        /// <returns>=true:正常、=false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>受信するデータが自モデルが期待している情報か否かをチェックする</para>
        /// </remarks>
        protected override bool ReceiveConnectionCheck(ref McStructErrorInfo csErrorInf, long lInputDataNum, McReceiveCellDataIF[] csInputCellData)
        {
            bool bRtn = true;

            // 入力端子数のチェック
            if (lInputDataNum == 0)
            {
                csErrorInf.AddCheckErrorData(this.m_csElement.GetID(), UnitGraphDefine.M_KIND, "No Receive Port");
                bRtn = false;
            }
            foreach (var recievePort in csInputCellData)
            {
                if (recievePort.GetReceivePatternID().Equals(UnitGraphDefine.IN_RAINFALL_POINT_PTN))
                {
                    // ポイント時系列の場合

                }
                else if (recievePort.GetReceivePatternID().Equals(UnitGraphDefine.IN_RAINFALL_1D_PTN))
                {
                    // １次元時系列の場合
                    long lD1 = 0;
                    long lD2 = 0;
                    long lD3 = 0;
                    long lCellDataNum = 0;
                    long lCellDim = recievePort.GetDimension(ref lD1, ref lD2, ref lD3, ref lCellDataNum);

                    if (lCellDim == 1 && lD1 >= 1 && lCellDataNum >= 1)
                    {
                        //do nothing
                    }
                    else
                    {
                        // エラー 
                        bRtn = false;
                        csErrorInf.AddCheckErrorData(this.m_csElement.GetID(), UnitGraphDefine.M_KIND,
                            "Unexpected receive pattern. (Received from " + m_csInputData[0].GetUpperElementID().ToString() + ")");
                    }
                }

                // エラー (期待している情報ではない)
                else
                {
                    bRtn = false;
                    csErrorInf.AddCheckErrorData(this.m_csElement.GetID(), UnitGraphDefine.M_KIND,
                        "Unexpected receive pattern. (Received from " + m_csInputData[0].GetUpperElementID().ToString() + ")");
                }

            }

            return bRtn;
        }

        /// <summary><para>method outline:</para>
        /// <para>出力側の接続情報チェック</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>bool bRtn = SendConnectionCheck(ref csErrorInf)</para>
        /// </example>
        /// <param name="csErrorInf">エラー出力</param>
        /// <param name="lOutputDataNum">出力情報数</param>
        /// <param name="csOutputCellData">出力情報配列</param>
        /// <returns>=true:正常、=false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>送信端子に設定されている伝送データが自モデルが期待している情報か否かをチェックする</para>
        /// </remarks>
        protected override bool SendConnectionCheck(ref McStructErrorInfo csErrorInf, long lOutputDataNum, McSendCellDataIF[] csOutputCellData)
        {
            bool bRtn = true;

            if (lOutputDataNum == 0)
            {
                // 警告（エラーではない）
                csErrorInf.AddCheckWarningData(this.GetID(), UnitGraphDefine.M_KIND, "No send port.");
            }
            foreach (var sendPort in csOutputCellData)
            {
                if (sendPort.GetSendPatternID().Equals(UnitGraphDefine.OUT_DISCHARGE_POINT_PTN))
                {
                    //ポイント時系列の場合

                }
                else if (sendPort.GetSendPatternID().Equals(UnitGraphDefine.OUT_DISCHARGE_1D_PTN))
                {
                    // 1次元時系列の場合
                    long lD1 = 0;
                    long lD2 = 0;
                    long lD3 = 0;
                    long lCellDataNum = 0;
                    long lCellDim = sendPort.GetDimension(ref lD1, ref lD2, ref lD3, ref lCellDataNum);

                    if (lCellDim == 1 && lD1 >= 1 && lCellDataNum >= 1)
                    {
                        // do nothing
                    }
                    else
                    {
                        // エラー 
                        bRtn = false;
                        csErrorInf.AddCheckErrorData(this.m_csElement.GetID(), UnitGraphDefine.M_KIND,
                            "Unexpected receive pattern. (Received from " + m_csInputData[0].GetUpperElementID().ToString() + ")");
                    }
                }
                else
                {
                    // エラー (期待している情報ではない)
                    bRtn = false;
                    csErrorInf.AddCheckErrorData(this.GetID(), UnitGraphDefine.M_KIND,
                        "Unexpected send pattern. (Send to " + sendPort.GetLowerElementID().ToString() + ")");
                }
            }

            return bRtn;
        }


        /// <summary><para>method outline:</para>
        /// <para>モデルを初期化する</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>Initialize(csInitialData)</para>
        /// </example>
        /// <param name="csInitialData">初期化設定情報</param>
        /// <param name="lInputDataNum">入力情報数</param>
        /// <param name="csInputCellData">演算に必要な入力情報配列</param>
        /// <returns>無し</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        protected override bool Initialize(ref McPropertyInfoRoot csInitialData, long lInputDataNum, ref McReceiveCellDataIF[] csInputCellData)
        {
            var c = m_csCalInfo as UnitGraphCalInfo;

            //シミュレーション開始期間をセット

            //格納用変数初期化
            //do nothing

            //パラメタ生成
            c.m_PrecipArray = new Queue<double>(UnitGraphCalInfo.RecordSize);
            Debug.Assert(c.m_PrecipArray.Count == 0);

            //変数正常性のチェック
            McStructErrorInfo csErrorInf = null;
            //HACK: to show error message to user, CheckParameters() is moved from SetProperty()
            //return CheckVariables(ref csErrorInf);
            return CheckParameters(ref csErrorInf) && CheckVariables(ref csErrorInf);
        }

        /// <summary><para>method outline:</para>
        /// <para>モデルを初期化する</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>OnlineInitialize(csInitialData)</para>
        /// </example>
        /// <param name="csInitialData">初期化設定情報</param>
        /// <param name="lInputDataNum">入力情報数</param>
        /// <param name="csInputCellData">演算に必要な入力情報配列</param>
        /// <returns>bool</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        protected override bool OnlineInitialize(ref McPropertyInfoRoot csInitialData, long lInputDataNum, ref McReceiveCellDataIF[] csInputCellData)
        {
            // 
            return true;
        }
        //=======================
        // 演算実行処理関連メソッド
        //=======================

        /// <summary><para>method outline:</para>
        /// <para>モデル演算</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>long lRtn = Calculate(ref csInputCellData)</para>
        /// </example>
        /// <param name="lInputDataNum">入力情報数</param>
        /// <param name="csInputCellData">演算に必要な入力情報配列</param>
        /// <returns>=0:正常 -1:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        protected override long Calculate(long lInputDataNum, ref McReceiveCellDataIF[] csInputCellData)
        {

            var c = m_csCalInfo as UnitGraphCalInfo;
            var timeStepInH = c.m_TimeStep / 3600.0; //時間ステップ, 単位[h]
            double rainfall = 0; //降雨量, 時間積分値(時間c.m_TimeStepの間), 単位[mm/h]

            //降雨量の受信
            foreach (var recievePort in csInputCellData)
            {
                if (recievePort.GetReceivePatternID().Equals(UnitGraphDefine.IN_RAINFALL_POINT_PTN))
                {
                    // ポイント時系列の場合
                    var cellId = recievePort.GetInterpolatedCell(0);  // セル配列の０番目のセルIDを取得
                    rainfall += recievePort.Data(cellId, 0);
                }
                else if (recievePort.GetReceivePatternID().Equals(UnitGraphDefine.IN_RAINFALL_1D_PTN))
                {
                    // １次元時系列の場合
                    var cellId = recievePort.GetInterpolatedCellD1();  // セル配列の０番目のセルIDを取得
                    rainfall += recievePort.Data(cellId[0], 0);
                }
            }
            if (c.m_PrecipArray.Count == UnitGraphCalInfo.RecordSize)
            {
                c.m_PrecipArray.Dequeue();
            }
            c.m_PrecipArray.Enqueue(rainfall);

            //降雨量時系列と流出関数を掛け合わせる
            double q = 0;
            Debug.Assert(c.m_PrecipArray.Count <= UnitGraphCalInfo.RecordSize);
            int i = c.m_PrecipArray.Count - 1;
            foreach (var r in c.m_PrecipArray)
            {
                q += r * c.m_RunoffFunction[i];
                i--;
            }
            q *= timeStepInH;

            //流域からの流量（合計値）を算出, 単位[m3/s]
            //単位に注意! q[mm/sec], Area[km2]
            c.m_DischargeTotal = q * c.m_Area;

            //if (m_csSimTime.GetTime() == m_csStartTime.GetTime())
            //{
            //    Debug.Print("t, c.m_DischargeTotal");
            //}
            //Debug.Print("{0},{1}", m_csSimTime.GetTime(), c.m_DischargeTotal);
            //foreach (var r in c.m_PrecipArray)
            //{
            //    Debug.Print("{0}", r);
            //}

            return 0;
        }


        /// <summary><para>method outline:</para>
        /// <para>モデル演算結果を外部のエレメントに対して公開する</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>long lRtn = DataFusion( lOutputDataNum, ref csOutputCellData)</para>
        /// </example>
        /// <param name="lOutputDataNum">出力情報数</param>
        /// <param name="csOutputCellData">出力情報配列</param>
        /// <returns>=0:正常 -1:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        protected override long DataFusion(long lOutputDataNum, ref McSendCellDataIF[] csOutputCellData)
        {
            var c = m_csCalInfo as UnitGraphCalInfo;

            // 出力
            foreach (var sendPort in csOutputCellData)
            {
                var cell = sendPort.PrepareSendCellD1();
                cell[0].m_dData[0] = c.m_DischargeTotal;
            }
            return 0;
        }


        //====================
        // その他必要なメソッド
        //====================

        /// <summary><para>method outline:</para>
        /// <para>プロパティ情報を設定する</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>SetProperty(csCellMdlPropertyInfo)</para>
        /// </example>
        /// <param name="csCellMdlPropertyInfo">セル型プロパティ情報</param>
        /// <returns>無し</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        public override bool SetProperty(McCellModelPropertyIF csCellMdlPropertyInfo)
        {
            // 使用しやすいようにキャストしておく
            var c = m_csCalInfo as UnitGraphCalInfo;

            // プロパティ設定
            var prop = csCellMdlPropertyInfo as McCellModelPropertyInfo;
            if (prop != null)
            {
                // 演算ステップ時刻設定
                m_csDltTime = new HySTime(prop.GetStepTime());

                // 各種係数設定
                c.m_TimeStep = prop.GetStepTime();

                prop.GetInfo("m_Area", ref c.m_Area);
                for (int i = 0; i < UnitGraphCalInfo.RecordSize; i++)
                {
                    var paramName = string.Format("m_RunoffFunc_{0}", i);
                    prop.GetInfo(paramName, ref c.m_RunoffFunction[i]);
                }

                //HACK: to show error message to user, CheckParameters() is moved to Initialize()
                //McStructErrorInfo csErrorInfo = null;
                //return CheckParameters(ref csErrorInf);
                return true;
            }
            else
            {

                return false;
            }
        }

        /// <summary><para>method outline:</para>
        /// <para>演算実行中断中のプロパティ等情報設定</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>bool bRtn = SetOnlineProperty(csCellMdlPropertyInfo)</para>
        /// </example>
        /// <param name="csCellMdlPropertyInfo">プロパティ情報</param>
        /// <returns>無し</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>演算中断中に　プロパティ情報を変更した場合にコールされる</para>
        /// </remarks>
        public override bool SetOnlineProperty(McCellModelPropertyIF csCellMdlPropertyInfo)
        {
            bool bRtn = false;
            // プロパティ設定
            McCellModelPropertyInfo csPrptyInfo = csCellMdlPropertyInfo as McCellModelPropertyInfo;
            if (csPrptyInfo != null)
            {
                bRtn = true;
            }
            return bRtn;
        }

        //private 
        private bool CheckVariables(ref McStructErrorInfo csErrorInfo)
        {
            return true;
        }
        private bool CheckParameters(ref McStructErrorInfo csErrorInfo)
        {
            var c = m_csCalInfo as UnitGraphCalInfo;
            bool ret = true;


            if (c.m_Area < 0)
            {
                if (csErrorInfo != null)
                    csErrorInfo.AddCheckErrorData(this.m_csElement.GetID(), UnitGraphDefine.M_KIND, Properties.Resources.STRING_AREA_OUT_OF_RANGE);
                McLog.DebugOut(GetSimulationTime(), GetID(), McLog.SYSTEM, " SetProperty Exception Catch", " in (" + this.ToString() + ") :: " + Properties.Resources.STRING_AREA_OUT_OF_RANGE);
                ret = false;
            }

            return ret;
        }


        /// <summary>
        /// <para>流出関数の生成</para>
        /// <para>ただし、流出関数として「高瀬の式」を利用</para>
        /// <para>参考文献: 高瀬信忠, 河川工学入門, 森北出版株式会社, pp.64-68, eq. 2.20, 2003.</para>
        /// <example>
        /// <code>
        /// double tg = 3; //[h]時間遅れ
        /// int n = 1; //流域形状パラメータ(通常、1 or 2)
        /// double timeStep = 1; //[sec]
        /// int recordSize = 20;
        /// GenerateRunoffFunction(1 / tg, n, timeStep / 3600.0, recordSize);
        /// </code>
        /// </example> 
        /// </summary>
        public static double[] GenerateRunoffFunction(double alpha, int n, double dt, int arraySize)
        {
            var func = new double[arraySize];
            for (int i = 0; i < arraySize; i++)
            {
                func[i] = 1.0 / 3.6 * Math.Pow(alpha, n + 1) / Factorial(n) * Math.Pow(dt * i, n) * Math.Exp(-alpha * dt * i);
            }
            return func;
        }

        private static int Factorial(int i)
        {
            switch (i)
            {
                case 0: return 1;
                case 1: return 1;
                case 2: return 2;
                case 3: return 6;
                case 4: return 24;
                case 5: return 120;
                case 6: return 720;
                case 7: return 5040;
                case 8: return 40320;
                case 9: return 362880;
                case 10: return 3628800;
                default: throw new ArgumentOutOfRangeException();
            }
        }
    }

}
