﻿// <summary>ソースコード：モデル要素演算データクラス</summary>
// <author>CommonMP</author>

using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;

using CommonMP.HYSSOP.Interface.HSData;
using CommonMP.HYSSOP.Interface.HSViewer;
using CommonMP.HYSSOP.CoreImpl.HSData;
using CommonMP.HYSSOP.CoreImpl.HSTools;

using CommonMP.HYMCO.Interface.Model;
using CommonMP.HYMCO.Interface.Data;
using CommonMP.HYMCO.Interface.Controller;

namespace jp.or.jccajsce.UnitGraph
{
    /// <summary><para>class outline:</para>
    /// <para>モデル要素内演算中データクラス</para>
    /// </summary>
    /// <remarks><para>history:</para>
    /// <para>[CommonMP][ver 1.0.0][2010/01][新規作成]</para>
    /// <para>remarks</para>
    /// <para>演算中の状態、初期条件等を保持するデータクラス</para>
    /// </remarks>
    [Serializable]
    public class UnitGraphCalInfo : McCalInfo
    {
        /// <summary>流出関数のレコード数</summary>
        public static int RecordSize = 48;

        // parameters
        /// <summary>流域面積 [km2]</summary>
        public double m_Area = 1;
        /// <summary>時間ステップ [s]</summary>
        public double m_TimeStep = 0;
        /// <summary>流出関数の形状 []</summary>
        public double[] m_RunoffFunction = new double[RecordSize];
        /// <summary>降雨時系列 [mm/h]</summary>
        public Queue<double> m_PrecipArray; 

        // variables
        /// <summary>流域からの流量 [m3/s]</summary>
        public double m_DischargeTotal = 0;

        /// <summary><para>method outline:</para>
        /// <para>自己複製</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para> Clone( ) </para>
        /// </example>
        /// <param name="">無し</param>
        /// <returns>同一内容で　別インスタンス</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>収束演算時、特定の時刻に状態を戻す場合等に使用する</para>
        /// </remarks>
        public override McCalInfo Clone()
        {
            var csRtn = new UnitGraphCalInfo();
            csRtn.CopyInfo(this);
            return csRtn;
        }

        /// <summary><para>method outline:</para>
        /// <para>引数で与えられた情報を自分にコピーを行う</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para> bool bRtn = CopyInfo(csOrgInfo) </para>
        /// </example>
        /// <param name="csOrgInfo">コピー元情報</param>
        /// <returns> bool true :正常 , false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        public override bool CopyInfo(McCalInfo csOrgInfo)
        {
            var c = (UnitGraphCalInfo)csOrgInfo;

            // parameters
            m_Area = c.m_Area;
            m_TimeStep = c.m_TimeStep;
            //deep copy
            c.m_RunoffFunction.CopyTo(m_RunoffFunction, RecordSize);
            m_PrecipArray.Clear();
            foreach (var elem in c.m_PrecipArray)
            {
                m_PrecipArray.Enqueue(elem);
            }

            // variables
            m_DischargeTotal = c.m_DischargeTotal;

            return true;
        }

    }

}
